"""Tests for the results module."""
from alglab.results import Results
import alglab.experiment
from sklearn.cluster import KMeans, SpectralClustering
import numpy as np


def test_plots():
    # Run a simple experiment
    results = Results("results/results.csv")
    assert results.num_runs == 2

    results.line_plot('noise', 'running_time_s')


def test_plots_multiple_parameters():
    noise_parameters = np.linspace(0, 1, 5)
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={'n_clusters': 2,
                    'dataset.noise': noise_parameters,
                    'dataset.n': np.linspace(100, 1000, 3).astype(int)},
        evaluators=[alglab.evaluation.adjusted_rand_index]
    )
    results = experiments.run_all()
    results.line_plot('n', 'total_running_time_s', fixed_parameters={'noise': noise_parameters[0]})





"""Classes and method for processing experimental results."""
import pandas as pd
import matplotlib.pyplot as plt
from typing import List
import numpy as np


class Results(object):

    def __init__(self, results_filename):
        """Load results from a csv file."""
        self.results_df = pd.read_csv(results_filename, skipinitialspace=True)
        self.algorithm_names = self.results_df['algorithm'].unique()
        self.num_runs = self.results_df['run_id'].max()
        self.stats_df = self.create_averaged_stats()

    def create_averaged_stats(self):
        """
        Create averages and error bars from multiple runs.
        """
        alg_dfs = {alg_name: self.results_df.loc[self.results_df['algorithm'] == alg_name] for alg_name in
                  self.algorithm_names}
        stats_df = pd.DataFrame()
        to_group_by = ['experiment_id']
        if 'iter' in self.results_df.columns:
            to_group_by.append('iter')
        for alg_name, alg_df in alg_dfs.items():
            mean_df = alg_df.groupby(to_group_by).mean(numeric_only=True)
            mean_df = mean_df.drop(['trial_id', 'run_id'], axis=1).add_prefix('_mean_')
            sem_df = alg_df.groupby(to_group_by).sem(numeric_only=True)
            sem_df = sem_df.drop(['trial_id', 'run_id'], axis=1).add_prefix('_sem_')
            sem_df['algorithm'] = alg_name
            sem_df = sem_df.join(mean_df)
            stats_df = pd.concat([stats_df, sem_df])
        return stats_df

    def column_names(self) -> List[str]:
        return self.results_df.columns.values.tolist()

    def line_plot(self, x_col, y_col, filename=None,
                  ignore_algorithms=None,
                  fixed_parameters=None,
                  algorithm_names=None,
                  x_range=None,
                  y_label=None,
                  x_label=None,
                  show_legend=True,):
        """Plot one column of the dataframe against another."""
        if ignore_algorithms is None:
            ignore_algorithms = []
        if fixed_parameters is None:
            fixed_parameters = {}
        if algorithm_names is None:
            algorithm_names = {}

        if x_col not in ['iter']:
            x_col = f"_mean_{x_col}"

        # Filter all the results by the provided fixed parameters
        results_to_plot = self.stats_df
        for param, val in fixed_parameters.items():
            results_to_plot = results_to_plot[results_to_plot[f'_mean_{param}'] == val]

        fig, ax = plt.subplots(figsize=(4, 3))
        plt.xlabel(x_col)
        plt.ylabel(y_col)
        plt.grid(True)

        for alg_name in self.algorithm_names:
            if alg_name not in ignore_algorithms:
                this_alg_name = algorithm_names[alg_name] if alg_name in algorithm_names else alg_name
                this_alg_results = results_to_plot[(results_to_plot['algorithm'] == alg_name)]
                plt.plot(this_alg_results[x_col],
                         this_alg_results[f"_mean_{y_col}"],
                         linewidth=3,
                         label=this_alg_name)
                if self.num_runs > 1:
                    plt.fill_between(this_alg_results[x_col],
                                     this_alg_results[f"_mean_{y_col}"] - this_alg_results[f"_sem_{y_col}"],
                                     this_alg_results[f"_mean_{y_col}"] + this_alg_results[f"_sem_{y_col}"],
                                     alpha=0.2)

        # Set the limits of the plot if specified
        if x_range is not None:
            ax.set_xlim(x_range[0], x_range[1])

        if x_label is not None:
            ax.set_xlabel(x_label)
        if y_label is not None:
            ax.set_ylabel(y_label)

        if show_legend:
            plt.legend()
        if filename:
            plt.savefig(filename, format="pdf", bbox_inches="tight")
        plt.show()

    def bar_plot(self, x_col, y_col, x_vals, filename=None,
                 ignore_algorithms=None,
                 fixed_parameters=None):
        if ignore_algorithms is None:
            ignore_algorithms = []

        if fixed_parameters is None:
            fixed_parameters = {}

        # Filter all the results by the provided fixed parameters
        results_to_plot = self.stats_df
        for param, val in fixed_parameters.items():
            results_to_plot = results_to_plot[results_to_plot[param] == val]

        x = np.arange(len(x_vals))  # the label locations
        width = 0.75 / (len(self.algorithm_names) - len(ignore_algorithms))  # the width of the bars

        fig, ax = plt.subplots(layout='constrained')

        multiplier = 0
        for alg_name in self.algorithm_names:
            if alg_name not in ignore_algorithms:
                offset = width * multiplier
                alg_results = results_to_plot[(self.stats_df['algorithm'] == alg_name)]
                y_vals = []
                for x_val in x_vals:
                    y_vals.append(alg_results[alg_results[f"_mean_{x_col}"] == x_val][f"_mean_{y_col}"].values[0])
                rects = ax.bar(x + offset, y_vals, width, label=alg_name)
                ax.bar_label(rects, padding=3)
                multiplier += 1

        plt.legend()
        if filename:
            plt.savefig(filename, format="pdf", bbox_inches="tight")
        plt.tight_layout()
        plt.show()




"""Tests for the experiment module."""
from sklearn.cluster import KMeans, SpectralClustering
from sklearn.mixture import GaussianMixture
import numpy as np
import pytest
import alglab

class DynamicKMeans(object):

    def __init__(self, k=10):
        self.current_data = []
        self.k = k

    def add_point(self, new_point: np.ndarray):
        self.current_data.append(new_point.tolist())

    def predict(self):
        sklearn_km = KMeans(n_clusters=self.k)
        sklearn_km.fit(self.current_data)
        return sklearn_km.labels_


class DynamicSC(object):

    def __init__(self, k=10):
        self.current_data = []
        self.k = k

    def add_point(self, new_point: np.ndarray):
        self.current_data.append(new_point.tolist())

    def predict(self):
        sklearn_km = SpectralClustering(n_clusters=self.k)
        sklearn_km.fit(self.current_data)
        return sklearn_km.labels_


def test_experimental_suite():
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "KMeans.n_clusters": 2,
            "SpectralClustering.n_clusters": 2,
            "dataset.n": 1000,
            "dataset.noise": np.linspace(0, 1, 5),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index]
        )
    experiments.run_all()


def test_multiple_runs():
    # Test the experimental suite class as it's intended to be used.
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "KMeans.n_clusters": 2,
            "SpectralClustering.n_clusters": 2,
            "dataset.n": 1000,
            "dataset.noise": np.linspace(0, 1, 5),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index],
        num_runs=2
    )
    assert experiments.num_trials == 20
    experiments.run_all()


def test_memory_measurements():
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "KMeans.n_clusters": 2,
            "SpectralClustering.n_clusters": 2,
            "dataset.n": np.linspace(1000, 3000, 3),
            "dataset.noise": 0.5,
        },
        evaluators=[alglab.evaluation.adjusted_rand_index],
        num_runs=1,
    )
    results = experiments.run_all()
    results.line_plot('n', 'memory_usage_mib')


def test_dynamic_params():
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "KMeans.n_clusters": 2,
            "SpectralClustering.n_clusters": [(lambda p: int(p['n'] / 100)), 2],
            "dataset.noise": 0.1,
            "dataset.n": np.linspace(100, 1000, 5).astype(int),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index]
    )
    experiments.run_all()


def test_simple_configuration():
    experiments = alglab.experiment.ExperimentalSuite([KMeans, SpectralClustering],
                                                      alglab.dataset.TwoMoonsDataset,
                                                      alglab.experiment.StaticClusteringSchedule,
                                                      "results/twomoonsresults.csv")
    experiments.run_all()


def test_simple_with_custom_evaluator():
    def const_evaluator(expected_output, alg_output):
        return 5

    experiments = alglab.experiment.ExperimentalSuite([KMeans, SpectralClustering],
                                                      alglab.dataset.TwoMoonsDataset,
                                                      alglab.experiment.StaticClusteringSchedule,
                                                      "results/twomoonsresults.csv",
                                                      evaluators=[const_evaluator])
    experiments.run_all()


def test_wrong_alg_name():
    algs = [KMeans, SpectralClustering]

    with pytest.raises(ValueError, match='algorithm'):
        experiments = alglab.experiment.ExperimentalSuite(
            algs,
            alglab.dataset.TwoMoonsDataset,
            alglab.experiment.StaticClusteringSchedule,
            "results/twomoonsresults.csv",
            parameters={
                "spectral_clustering.k": 2,
                "dataset.n": 1000,
                "dataset.noise": np.linspace(0, 1, 5),
            },
            evaluators=[alglab.evaluation.adjusted_rand_index]
        )


def test_multi_step_algs():
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, GaussianMixture],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.FitPredictSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "KMeans.n_clusters": 2,
            "GaussianMixture.n_components": 2,
            "dataset.n": np.linspace(100, 1000, 6),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index],
    )

    results = experiments.run_all()
    results.line_plot("n", "fit_total_running_time_s")
    results.line_plot("n", "predict_total_running_time_s")
    results.line_plot("n", "total_running_time_s")


def test_dynamic_algs():
    experiments = alglab.experiment.ExperimentalSuite([DynamicKMeans, DynamicSC],
                                                      alglab.dataset.TwoMoonsDataset,
                                                      alglab.experiment.DynamicClusteringSchedule,
                                                      "results/dynamictwomoonsresults.csv",
                                                      parameters={
                                                          "DynamicKMeans.k": 2,
                                                          "DynamicSC.k": 2,
                                                          "schedule.batch_size": 100,
                                                      },
                                                      evaluators=[alglab.evaluation.adjusted_rand_index,
                                                                  alglab.evaluation.dataset_size])
    results = experiments.run_all()

    # Plot the per-iteration running times for each algorithm
    results.line_plot("dataset_size", "total_running_time_s")




import pytest
import numpy as np
import stag.graph
import alglab.dataset
import matplotlib.pyplot as plt


def test_no_dataset():
    # Check that initialising the dataset does not throw any error.
    _ = alglab.dataset.NoDataset()

    # Initialising with keyword arguments should give an error.
    with pytest.raises(Exception, match="keyword"):
        _ = alglab.dataset.NoDataset(eps=0.1)


def test_graph_dataset():
    graph = stag.graph.cycle_graph(100)
    dataset = alglab.dataset.GraphDataset(graph=graph)
    assert dataset.n == 100


def test_sbm_dataset():
    dataset = alglab.dataset.SBMDataset(n=100, k=2, p=0.5, q=0.1)
    assert dataset.n == 100

    with pytest.raises(Exception, match="p must be between 0 and 1"):
        _ = alglab.dataset.SBMDataset(n=100, k=2, p=-0.5, q=0.1)


def test_sbm_floats():
    # Check that we can initialise an SBM dataset with float arguments
    dataset = alglab.dataset.SBMDataset(n=100.0, k=2.1, p=0.5, q=0.1)
    assert dataset.n == 100


def test_pointcloud_dataset():
    data = np.asarray([[1, 1], [4, 2], [2, 3]])
    _ = alglab.dataset.PointCloudDataset(data=data)


def test_pointcloud_dataset_plot(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    data = np.asarray([[1, 1], [4, 2], [2, 3]])
    gt_labels = [0, 0, 1]
    dataset = alglab.dataset.PointCloudDataset(data=data, labels=gt_labels)
    dataset.plot_clusters(gt_labels)

    # Plotting a 3 dimensional dataset does not work
    with pytest.raises(Exception, match="two dimensions"):
        data = np.asarray([[1, 1, 1], [2, 3, 3]])
        gt_labels = [0, 1]
        dataset = alglab.dataset.PointCloudDataset(data=data, labels=gt_labels)
        dataset.plot_clusters(gt_labels)

    # If the labels array is not the correct length, raises an exception
    with pytest.raises(Exception, match="length"):
        data = np.asarray([[1, 1], [4, 2], [2, 3]])
        dataset = alglab.dataset.PointCloudDataset(data=data)
        labels = [1, 1]
        dataset.plot_clusters(labels)


def test_twomoons_dataset(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)

    # Check that we can create the two moons dataset
    dataset = alglab.dataset.TwoMoonsDataset(n=200, noise=0.02)

    # Check that we can plot the dataset
    dataset.plot_clusters(dataset.gt_labels)


def test_blobs_dataset(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)

    # Check that we can create the two moons dataset
    dataset = alglab.dataset.BlobsDataset(n=200)

    # Check that we can plot the dataset
    dataset.plot_clusters(dataset.gt_labels)


def test_circles_dataset(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)

    # Check that we can create the two moons dataset
    dataset = alglab.dataset.CirclesDataset(n=200, noise=0.02)

    # Check that we can plot the dataset
    dataset.plot_clusters(dataset.gt_labels)


def test_knearest_neighbour_dataset(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)

    dataset = alglab.dataset.KnnGraphDataset(k=5, pointcloud_class=alglab.dataset.TwoMoonsDataset,
                                            n=200, noise=0.02)

    # Check that we can plot the dataset
    dataset.plot_clusters(dataset.gt_labels)

    # Check that we have a number of vertices
    assert dataset.n == 200
    assert dataset.graph.number_of_vertices() == 200


def test_openml_dataset():
    dataset = alglab.dataset.OpenMLDataset(name="iris")
    assert dataset.n == 150
    assert dataset.d == 4
    assert isinstance(dataset.gt_labels, np.ndarray)


def test_pca_and_scaling():
    dataset = alglab.dataset.OpenMLDataset(name="iris")
    assert dataset.n == 150
    assert dataset.d == 4

    dataset.apply_pca(2)
    dataset.apply_scaling()
    assert dataset.n == 150
    assert dataset.d == 2


def test_plot_dataset():
    dataset = alglab.dataset.TwoMoonsDataset()
    dataset.plot_data()


def test_plot_dimensions():
    dataset = alglab.dataset.OpenMLDataset(name="iris")
    dataset.plot_data(dimension_idxs=[1, 2])



"""Implementation of the Dataset object for use with algpy."""
from sklearn.datasets import make_moons, fetch_openml, make_blobs, make_circles
from sklearn.neighbors import kneighbors_graph
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import numpy as np
from abc import ABC, abstractmethod
import stag.graph
import stag.random
import matplotlib.pyplot as plt
from typing import Type, Dict, List, Tuple
import pandas as pd


class Dataset(ABC):

    def __init__(self, **kwargs):
        """Construct the dataset."""
        self.num_steps = 1

    def get_step(self, step_id):
        """Get the update which must happen at the given step.
        An update has the form (method_name, arguments, gt_output)."""
        return


class NoDataset(Dataset):
    """Use this when no dataset is needed to compare algorithms."""

    def __init__(self):
        super().__init__()
        self.num_steps = 0

    def __str__(self):
        return "NoDataset"


class ClusterableDataset(Dataset):
    """
    A dataset which may have ground truth clusters.
    """

    def __init__(self, labels, **kwargs):
        super().__init__(**kwargs)
        self.gt_labels = labels

    def cluster_ids(self):
        if self.gt_labels is None:
            return None
        else:
            return np.unique(self.gt_labels).tolist()

    def get_cluster(self, cluster_id: int):
        cluster = []
        for id, lab in enumerate(self.gt_labels):
            if lab == cluster_id:
                cluster.append(id)
        return cluster

    def get_step(self, step_id):
        return 'fit_predict', None, self.gt_labels


class GraphDataset(ClusterableDataset):
    """
    A dataset whose central data is a graph.
    """

    def __init__(self, graph: stag.graph.Graph = None, labels=None):
        """Initialise the dataset with a stag Graph. Optionally, provide ground truth
        labels for classification."""
        self.graph = graph
        self.n = 0 if graph is None else graph.number_of_vertices()
        ClusterableDataset.__init__(self, labels)

    def get_step(self, step_id):
        return 'fit_predict', self.graph, self.gt_labels


class SBMDataset(GraphDataset):
    """
    Create a graph dataset from a stochastic block model.
    """

    def __init__(self, n: int = 1000, k: int = 10, p: float = 0.5, q: float = 0.1):
        self.n = int(n)
        self.k = int(k)
        self.p = p
        self.q = q
        g = stag.random.sbm(self.n, self.k, p, q)
        labels = stag.random.sbm_gt_labels(self.n, self.k)
        GraphDataset.__init__(self, graph=g, labels=labels)


    def __repr__(self):
        return f"SBMDataset({self.n}, {self.k}, {self.p}, {self.q})"


class PointCloudDataset(ClusterableDataset):
    """
    The simplest form of dataset: the data consists of a point cloud in Euclidean space.
    This is represented internally by a numpy array.
    """

    def __init__(self, data: np.array = None, labels=None):
        """Initialise the dataset with a numpy array. Optionally, provide labels for classification."""
        self.data = np.array(data)
        self.n, self.d = data.shape
        self.k = -1 if labels is None else len(np.unique(np.asarray(labels)))
        ClusterableDataset.__init__(self, labels)

    def apply_pca(self, new_dimension: int):
        pca = PCA(n_components=new_dimension)
        self.data = pca.fit_transform(self.data)
        assert self.data.shape[1] == new_dimension
        self.d = new_dimension

    def apply_scaling(self):
        scaler = StandardScaler().fit(self.data)
        self.data = scaler.transform(self.data)

    def plot_clusters(self, labels, dimension_idxs=None):
        """
        If the data is two-dimensional, plot the data, colored according to the labels.

        If dimension_idxs is an array of length two, plot the data using the given dimension indices.
        """
        if (self.d != 2 and dimension_idxs is None) or (dimension_idxs is not None and len(dimension_idxs) != 2):
            raise ValueError("Cannot plot dataset: it has more than two dimensions.")

        if len(labels) != self.n:
            raise ValueError("Cannot plot dataset: labels length must match number of data points.")

        if dimension_idxs is None:
            dimension_idxs = [0, 1]

        labels = np.array(labels)

        # Plot the data points, colored by their cluster labels
        plt.figure(figsize=(10, 7))
        unique_labels = np.unique(labels)

        for label in unique_labels:
            cluster_data = self.data[labels == label]
            plt.scatter(cluster_data[:, dimension_idxs[0]], cluster_data[:, dimension_idxs[1]], label=f'Cluster {label}')

        plt.grid(True)
        plt.show()

    def plot_data(self, dimension_idxs=None):
        """
        If the data is two-dimensional, plot it.

        If dimension_idxs is an array of length two, plot the data using the given dimension indices.
        """
        labels = np.ones(self.n)
        self.plot_clusters(labels, dimension_idxs=dimension_idxs)

    def get_step(self, step_id):
        return 'fit_predict', self.data, self.gt_labels


class TwoMoonsDataset(PointCloudDataset):
    """The toy two moons dataset from sklearn."""

    def __init__(self, n=1000, noise=0.07):
        """Initialise the two moons dataset. Optionally, provide the number of points, n, and the noise parameter."""
        x, y = make_moons(n_samples=int(n), noise=noise)
        PointCloudDataset.__init__(self, data=x, labels=y)

    def __str__(self):
        return f"TwoMoonsDataset({self.n})"


class BlobsDataset(PointCloudDataset):
    """The toy blobs dataset from sklearn."""

    def __init__(self, n=1000, d=2, k=3):
        """Initialise the blobs dataset. Optionally, provide the number of points, dimensions, and clusters."""
        x, y = make_blobs(n_samples=n, n_features=d, centers=k)
        PointCloudDataset.__init__(self, data=x, labels=y)

    def __str__(self):
        return f"BlobsDataset({self.n}, {self.d}, {self.k})"


class CirclesDataset(PointCloudDataset):
    """The toy circles dataset from sklearn."""

    def __init__(self, n=1000, noise=0.07):
        """Initialise the circles dataset. Optionally, provide the number of points and the noise parameter."""
        x, y = make_circles(n_samples=int(n), noise=noise)
        PointCloudDataset.__init__(self, data=x, labels=y)

    def __str__(self):
        return f"CirclesDataset({self.n})"


class OpenMLDataset(PointCloudDataset):
    """Load pointcloud data from OpenML."""

    def __init__(self, **kwargs):
        """Initialise the dataset by downloading from openML. Accepts the same arguments as the
        sklearn fetch_openml method."""
        data_info = fetch_openml(**kwargs)
        if isinstance(data_info.data, pd.DataFrame):
            data_info.data = data_info.data.to_numpy()

        target = data_info.target
        if isinstance(target, pd.Series) or isinstance(target, pd.DataFrame):
            if isinstance(target.dtype, pd.CategoricalDtype):
                target = target.cat.codes.to_numpy()
            else:
                target = data_info.target.to_numpy()

        PointCloudDataset.__init__(self, data=data_info.data, labels=target)


class KnnGraphDataset(GraphDataset, PointCloudDataset):
    """A k-nearest neighbour graph dataset is both a point cloud and a graph dataset."""

    def __init__(self,
                 k: int = 10,
                 pointcloud_class: Type[PointCloudDataset] = PointCloudDataset,
                 **pointcloud_parameters):
        # Initialise this as a pointcloud dataset
        pointcloud_class.__init__(self, **pointcloud_parameters)

        # Create the k nearest neighbours graph and initialise as a graph dataset
        adj_non_symmetric = kneighbors_graph(self.data, k)
        g = stag.graph.Graph(adj_non_symmetric + adj_non_symmetric.transpose())
        GraphDataset.__init__(self, graph=g, labels=self.gt_labels)




"""
A simple example demonstrating how to use AlgLab to compare two clustering algorithms.
"""
from sklearn.cluster import KMeans, SpectralClustering
import numpy as np
import alglab


def main():
    # Configure the experiments. As well as the algorithms, we specify which dataset class to use,
    # and the parameters for the algorithms and dataset.
    #
    # The experiment schedule is given as the third argument and specifies which methods should be called on the
    # algorithm objects in order to run the experiment.
    #
    # We also specify any functions which should be used to evaluate the algorithms, and give a
    # filename in which to store the results.
    experiments = alglab.experiment.ExperimentalSuite(
        [KMeans, SpectralClustering],
        alglab.dataset.TwoMoonsDataset,
        alglab.experiment.StaticClusteringSchedule,
        "results/twomoonsresults.csv",
        parameters={
            "n_clusters": 2,
            "dataset.n": np.linspace(1000, 5000, 6).astype(int),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index],
    )

    # Run the experiments
    experiments.run_all()

    # Now, we can visualise the results
    results = alglab.results.Results("results/twomoonsresults.csv")
    results.line_plot("n", "total_running_time_s")
    results.line_plot("n", "adjusted_rand_index")


if __name__ == "__main__":
    main()



"""
Methods for evaluating the performance of an algorithm.
"""
from typing import Callable, Type, get_type_hints, Union, List
import numpy as np
import alglab
import stag.cluster
import stag.graph
import scipy.sparse.linalg
import inspect


def adjusted_rand_index(gt_labels, labels):
    # Remove any negative cluster ids
    next_cluster_id = np.max(labels) + 1
    mapping = {}
    indices_to_change = []
    for i, label in enumerate(labels):
        if label < 0:
            if label not in mapping:
                mapping[label] = next_cluster_id
                next_cluster_id += 1
            indices_to_change.append(i)
    for i in indices_to_change:
        labels[i] = mapping[labels[i]]

    if gt_labels is not None:
        return stag.cluster.adjusted_rand_index(gt_labels, labels)
    else:
        raise ValueError('No ground truth labels provided.')


def dataset_size(gt_labels, _):
    return len(gt_labels)


# -----------------------------------------------------------------------------
# Graph Evaluation
# -----------------------------------------------------------------------------

def num_vertices(_, graph: stag.graph.Graph):
    return graph.number_of_vertices()

def avg_degree(_, graph: stag.graph.Graph):
    return graph.average_degree()

def normalised_laplacian_second_eigenvalue(_, graph: stag.graph.Graph):
    lap = graph.normalised_laplacian().to_scipy()
    eigs, _ = scipy.sparse.linalg.eigsh(lap, which='SM', k=2)
    return eigs[1]



"""Classes and methods related to running experiments on algorithms and datasets."""
import time

from abc import ABC, abstractmethod

from typing import Dict, Type, List, Callable, Union
import itertools
from collections import OrderedDict

import psutil
import os
import threading
import numpy as np

import alglab.dataset
import alglab.results
import alglab.evaluation


def product_dict(**kwargs):
    keys = kwargs.keys()
    for instance in itertools.product(*kwargs.values()):
        yield dict(zip(keys, instance))


def resolve_parameters(fixed_parameters, varying_parameters):
    """
    Given a dictionary of fixed parameters and a dictionary of varying parameters, resolve the varying parameters
    which are defined by function.
    """
    resolved_parameters = {}
    dynamic_parameters = []
    for param_name, value in varying_parameters.items():
        if not callable(value):
            # This is a 'static' parameter
            resolved_parameters[param_name] = value
        else:
            # This is a dynamically defined parameter
            dynamic_parameters.append(param_name)

    # Resolve the dynamic parameters
    for param_name in dynamic_parameters:
        resolved_parameters[param_name] = varying_parameters[param_name](fixed_parameters | resolved_parameters)

    return resolved_parameters


class ExperimentSchedule(ABC):
    def __init__(self, dataset: alglab.dataset.Dataset, **kwargs):
        self.dataset = dataset

    @abstractmethod
    def schedule(self):
        pass

    @staticmethod
    @abstractmethod
    def all_method_names():
        pass


class StaticClusteringSchedule(ExperimentSchedule):
    def __init__(self, dataset: alglab.dataset.ClusterableDataset):
        super().__init__(dataset)

    def schedule(self):
        yield 'fit_predict', self.dataset.data, self.dataset.gt_labels

    @staticmethod
    def all_method_names():
        return ['fit_predict']


class FitPredictSchedule(ExperimentSchedule):
    def __init__(self, dataset: alglab.dataset.ClusterableDataset):
        super().__init__(dataset)

    def schedule(self):
        yield 'fit', self.dataset.data, None
        yield 'predict', self.dataset.data, self.dataset.gt_labels

    @staticmethod
    def all_method_names():
        return ['fit', 'predict']


class DynamicClusteringSchedule(ExperimentSchedule):
    def __init__(self, dataset: alglab.dataset.ClusterableDataset, batch_size=1000, stream_by_cluster=False):
        super().__init__(dataset)
        self.batch_size = batch_size
        self.stream_by_cluster = stream_by_cluster

    def schedule(self):
        if self.stream_by_cluster:
            # Stream in the data one cluster at a time
            ordered_indexes = []
            gt_labels = []
            for cluster_id in self.dataset.cluster_ids():
                for id in self.dataset.get_cluster(cluster_id):
                    ordered_indexes.append(id)
                    gt_labels.append(cluster_id)

            reordered_data = self.data[ordered_indexes, :]
        else:
            random_order = np.random.permutation(self.dataset.data.shape[0])
            reordered_data = self.dataset.data[random_order, :]
            gt_labels = self.dataset.gt_labels[random_order]

        current_n = 0
        num_since_last_query = 0
        while current_n < self.dataset.n:
            yield 'add_point', reordered_data[current_n, :], None
            num_since_last_query += 1
            current_n += 1

            if num_since_last_query >= self.batch_size:
                yield 'predict', None, gt_labels[:current_n]
                num_since_last_query = 0

        if num_since_last_query != 0:
            yield 'predict', None, gt_labels[:current_n]

    @staticmethod
    def all_method_names():
        return ['add_point', 'predict']


def monitor_memory(interval, stop_event, result_dict, base_memory):
    process = psutil.Process(os.getpid())
    peak_diff = 0
    while not stop_event.is_set():
        current = process.memory_info().rss
        diff = current - base_memory
        peak_diff = max(peak_diff, diff)
        time.sleep(interval)
    result_dict['peak_diff'] = peak_diff


class Experiment(object):

    def __init__(self, alg: Type,
                 schedule: ExperimentSchedule, params,
                 evaluators = None):
        """An experiment is a single instance of running an algorithm with a set of parameters according
        to some experiment schedule which indicates the order in which algorithm methods should be executed.
        The running time of the algorithm is measured by default. In addition to this, the evaluation_functions
        variable should contain a dictionary of methods which will be applied to the result of the algorithm.
        """
        self.alg = alg
        self.schedule = schedule
        self.params = params
        self.evaluators = evaluators

    def run(self):
        """Run the experiment."""
        # Set up the memory tracking
        this_process = psutil.Process(os.getpid())
        base_memory = this_process.memory_info().rss
        memory_dict = {}
        stop_event = threading.Event()

        # Create the algorithm object
        try:
            alg_obj = self.alg(**self.params)
        except TypeError as e:
            raise TypeError(f"Type error when initialising algorithm {self.alg.__name__}: {e.args[0]}")

        iter = 0
        results_dict = {}
        total_running_times = {}
        iter_running_time = 0
        total_running_time = 0
        for method_name, data, expected_result in self.schedule.schedule():
            # Run the algorithm and measure time and memory usage
            monitor_thread = threading.Thread(
                target=monitor_memory,
                args=(0.1, stop_event, memory_dict, base_memory)
            )
            monitor_thread.start()
            start_time = time.time()
            method_to_run = getattr(alg_obj, method_name)
            if data is None:
                alg_output = method_to_run()
            else:
                alg_output = method_to_run(data)
            end_time = time.time()
            stop_event.set()
            monitor_thread.join()
            peak_memory_bytes = memory_dict.get('peak_diff', 0)
            total_running_time += end_time - start_time
            iter_running_time += end_time - start_time

            iter_step_running_time_key = f'{method_name}_iter_running_time_s'
            iter_total_running_time_key = f'{method_name}_total_running_time_s'
            if iter_step_running_time_key not in results_dict:
                results_dict[iter_step_running_time_key] = end_time - start_time
            else:
                results_dict[iter_step_running_time_key] += end_time - start_time

            if iter_total_running_time_key not in total_running_times:
                total_running_times[iter_total_running_time_key] = end_time - start_time
            else:
                total_running_times[iter_total_running_time_key] += end_time - start_time

            if expected_result is not None:
                # This is the end of an iteration, and we will evaluate
                results_dict |= {
                    'iter': iter,
                    'iter_running_time_s': iter_running_time,
                    'total_running_time_s': total_running_time,
                    'memory_usage_mib': peak_memory_bytes / 1024 / 1024,
                }

                # Apply the evaluation functions
                if self.evaluators:
                    for evaluator in self.evaluators:
                        results_dict[evaluator.__name__] = evaluator(expected_result, alg_output)

                yield results_dict | total_running_times

                iter += 1
                results_dict = {}
                iter_running_time = 0


class ExperimentalSuite(object):

    def __init__(self,
                 algorithms: List[Type],
                 dataset: Union[Type[alglab.dataset.Dataset], alglab.dataset.Dataset],
                 schedule: Type[ExperimentSchedule],
                 results_filename: str,
                 num_runs: int = 1,
                 parameters: Dict = None,
                 evaluators: List[Callable] = None):
        """Run a suite of experiments while varying some parameters.

        Varying parameter dictionaries should have parameter names as keys and the values should be an iterable containing:
            - values to be used directly; or
            - functions, taking fixed and statically defined variable parameters and returning a parameter value
        """
        self.num_runs = num_runs

        if num_runs < 1:
            raise ValueError('num_runs must be greater than or equal to 1')

        self.algorithms = algorithms
        self.algorithm_names = [alg.__name__ for alg in self.algorithms]

        # Automatically populate the parameter dictionaries
        alg_fixed_params = {}
        alg_varying_params = {}
        dataset_fixed_params = {}
        dataset_varying_params = {}
        schedule_fixed_params = {}
        schedule_varying_params = {}

        if parameters is None:
            parameters = {}

        for param_name, param_value in parameters.items():
            alg_dataset_name = ""
            if "." in param_name:
                alg_dataset_name, parameter_name = param_name.split(".")
            else:
                parameter_name = param_name

            if alg_dataset_name == "dataset":
                try:
                    _ = iter(param_value)
                except TypeError:
                    dataset_fixed_params[parameter_name] = param_value
                else:
                    dataset_varying_params[parameter_name] = param_value
            elif alg_dataset_name == "schedule":
                try:
                    _ = iter(param_value)
                except TypeError:
                    schedule_fixed_params[parameter_name] = param_value
                else:
                    schedule_varying_params[parameter_name] = param_value
            elif alg_dataset_name != "":
                try:
                    _ = iter(param_value)
                except TypeError:
                    if alg_dataset_name not in alg_fixed_params:
                        alg_fixed_params[alg_dataset_name] = {}
                    alg_fixed_params[alg_dataset_name][parameter_name] = param_value
                else:
                    if alg_dataset_name not in alg_varying_params:
                        alg_varying_params[alg_dataset_name] = {}
                    alg_varying_params[alg_dataset_name][parameter_name] = param_value
            else:
                for alg_name in self.algorithm_names:
                    try:
                        _ = iter(param_value)
                    except TypeError:
                        if alg_name not in alg_fixed_params:
                            alg_fixed_params[alg_name] = {}
                        alg_fixed_params[alg_name][parameter_name] = param_value
                    else:
                        if alg_name not in alg_varying_params:
                            alg_varying_params[alg_name] = {}
                        alg_varying_params[alg_name][parameter_name] = param_value

        # Check that all the parameters make sense
        for alg in self.algorithms:
            alg_name = alg.__name__

            # Check that every algorithm has an entry in the params dictionary
            if alg_name not in alg_fixed_params:
                alg_fixed_params[alg_name] = {}
            if alg_name not in alg_varying_params:
                alg_varying_params[alg_name] = {}

            # Convert the parameter iterables to lists
            for param_name in alg_varying_params[alg_name].keys():
                alg_varying_params[alg_name][param_name] = list(alg_varying_params[alg_name][param_name])

        # Check that all configured parameters match some algorithm
        for alg in alg_fixed_params.keys():
            if alg not in self.algorithm_names:
                raise ValueError(f"Parameters configured for algorithm {alg} which does not exist.")
        for alg in alg_varying_params.keys():
            if alg not in self.algorithm_names:
                raise ValueError(f"Parameters configured for algorithm {alg} which does not exist.")

        #  Convert parameter iterables to lists
        for param_name in dataset_varying_params.keys():
            dataset_varying_params[param_name] = list(dataset_varying_params[param_name])

        self.alg_fixed_params = alg_fixed_params
        self.alg_varying_params = alg_varying_params
        self.dataset = None
        self.dataset_class = None
        if isinstance(dataset, alglab.dataset.Dataset):
            self.dataset = dataset
        else:
            self.dataset_class = dataset
        self.dataset_fixed_params = dataset_fixed_params
        self.dataset_varying_params = dataset_varying_params

        self.schedule_class = schedule
        self.schedule_fixed_params = schedule_fixed_params
        self.schedule_varying_params = schedule_varying_params

        self.evaluators = evaluators if evaluators is not None else []
        self.results_filename = results_filename

        self.results_columns = self.get_results_df_columns()

        # Compute the total number of experiments to run
        num_datasets = 1
        for param_name, values in self.dataset_varying_params.items():
            num_datasets *= len(values)

        num_schedules = 1
        for param_name, values in self.schedule_varying_params.items():
            num_schedules *= len(values)

        self.num_experiments = 0
        for alg_name in self.algorithm_names:
            num_experiments_this_alg = 1
            for param_name, values in self.alg_varying_params[alg_name].items():
                num_experiments_this_alg *= len(values)
            self.num_experiments += num_experiments_this_alg * num_datasets * num_schedules
        self.num_trials = self.num_experiments * self.num_runs

        self.results = None

    def get_results_df_columns(self):
        """Create a list of all the columns in the results file and dataframe."""
        columns = ['trial_id', 'experiment_id', 'run_id', 'algorithm', 'iter', 'iter_running_time_s', 'total_running_time_s', 'memory_usage_mib']
        for param_name in self.dataset_fixed_params.keys():
            columns.append(param_name)
        for param_name in self.dataset_varying_params.keys():
            columns.append(param_name)
        for param_name in self.schedule_fixed_params.keys():
            columns.append(param_name)
        for param_name in self.schedule_varying_params.keys():
            columns.append(param_name)

        for alg_name in self.algorithm_names:
            for param_name in self.alg_fixed_params[alg_name].keys():
                columns.append(param_name)
            for param_name in self.alg_varying_params[alg_name].keys():
                columns.append(param_name)

        for method_name in self.schedule_class.all_method_names():
            columns.append(f'{method_name}_iter_running_time_s')
            columns.append(f'{method_name}_total_running_time_s')

        for evaluator in self.evaluators:
            columns.append(evaluator.__name__)

        return list(OrderedDict.fromkeys(columns))

    def run_all(self, append_results=False) -> alglab.results.Results:
        """Run all the experiments in this suite."""

        # If we are appending the results, make sure that the header of the results file already matches the
        # header we would have written.
        if append_results:
            existing_results = alglab.results.Results(self.results_filename)
            if existing_results.column_names() != self.results_columns:
                raise ValueError("Cannot append results file: column names do not match.")
            true_trial_number = existing_results.results_df.iloc[-1]["trial_id"] + 1
            base_experiment_number = existing_results.results_df.iloc[-1]["experiment_id"] + 1
        else:
            true_trial_number = 1
            base_experiment_number = 1

        reported_trial_number = 1

        file_access_string = 'a' if append_results else 'w'

        with open(self.results_filename, file_access_string) as results_file:
            # Write the header line of the results file
            if not append_results:
                results_file.write(", ".join(self.results_columns))
                results_file.write("\n")

            for run in range(1, self.num_runs + 1):
                experiment_number = base_experiment_number
                for dataset_params in product_dict(**self.dataset_varying_params):
                    resolved_varying_dataset_params = resolve_parameters(self.dataset_fixed_params, dataset_params)
                    full_dataset_params = self.dataset_fixed_params | resolved_varying_dataset_params
                    dataset = self.dataset
                    if self.dataset is None:
                        dataset = self.dataset_class(**full_dataset_params)

                    for schedule_params in product_dict(**self.schedule_varying_params):
                        resolved_schedule_params = resolve_parameters(self.schedule_fixed_params, schedule_params)
                        full_schedule_params = self.schedule_fixed_params | resolved_schedule_params
                        schedule = self.schedule_class(dataset, **full_schedule_params)

                        for alg in self.algorithms:
                            alg_name = alg.__name__
                            for alg_params in product_dict(**self.alg_varying_params[alg_name]):
                                resolved_varying_alg_params = resolve_parameters(full_dataset_params | self.alg_fixed_params[alg_name], alg_params)
                                full_alg_params = self.alg_fixed_params[alg_name] | resolved_varying_alg_params
                                print(f"Trial {reported_trial_number} / {self.num_trials}: {alg_name} on {dataset} with parameters {full_alg_params}")
                                this_experiment = Experiment(alg, schedule, full_alg_params, self.evaluators)

                                for result in this_experiment.run():
                                    this_result = result | full_dataset_params | full_alg_params | full_schedule_params | \
                                                  {'algorithm': alg_name, 'trial_id': true_trial_number,
                                                   'experiment_id': experiment_number, 'run_id': run}
                                    results_file.write(", ".join([str(this_result[col]) if col in this_result else '' for col in self.results_columns]))
                                    results_file.write("\n")
                                    results_file.flush()

                                true_trial_number += 1
                                reported_trial_number += 1
                                experiment_number += 1

        # Create a dataframe from the results
        self.results = alglab.results.Results(self.results_filename)
        return self.results
